# GNIRS data reduction script
#
# This data processing was done for data quality assessment by Gemini staff.
# The processing is not designed to give the best possible result.  The user
# of these data is encouraged to use the provided "final" images, and the
# processing steps described here, only as a guide and to re-reduce the
# data carefully to obtain the best possible reduction.
#
# Observation UT date: 2004 Jun 15
# Data processor: B. Rodgers
# Data filename prefix: S20040615S
#
# Comments from data processor:
#
#  This is an example of a basic calibration reduction for longslit/32 
#   (R~1700) dataset with central wavelength=2.2microns.  The data consist 
#   of GCAL flats, an arc and a telluric standard.  
#
#   The obslog entries are included below just before the processing steps.
#

# This line points to the raw data 
set rawdir=/net/veronica/staging/gnirs/2004jun15/

# Load required tasks/packages 
gemini
gemtools
gnirs

#-----------setup for gnirs------------
#set default parameters
#use gnirs defaults except where specified on command lines below
unlearn gemini
unlearn gemtools
unlearn gnirs

#set logfile
delete GS-CAL20040615.log
gnirs.logfile="GS-CAL20040615.log"
#set database dir
gnirs.database="GS-CAL20040615_database"
#to start from scratch, delete database entries if present
delete GS-CAL20040615_database/*

#load gnirs header keywords
nsheaders gnirs
#set display
set stdimage=imt1024

###Start processing

#### Obslog entries for SV-8 Calibrations
#
#GS-CAL20040615-13  145-152 02:15   HIP60235     abba     K     0.3        32/2.208     SB/mir     5/1/2         telluric for SV-8; AM=1.23
#GS-CAL20040615-16  184-194 03:50   GCAL IR      none     K     0.3        32/2.208     SB/mir     0.5/1/2       flats for SV-8
#GS-CAL20040615-17  195-196 03:52   GCAL Ar      none     K     0.3        32/2.208     SB/mir     3.0/1/4       arcs for SV-8
###

#-----Create lists-----
del flat.lis
gemlist S20040615S 184-194 > flat.lis
del arc.lis
gemlist S20040615S 195 > arc.lis
del telluric.lis
gemlist S20040615S 145-152 > telluric.lis

#------prepare files----------------
#do this in a single batch, because we find the MDF shift by cross-correlation
#once and then use the same value for all files
imdelete n@flat.lis,n@arc.lis,n@telluric.lis
nsprepare @flat.lis,@arc.lis,@telluric.lis shiftx=INDEF shifty=INDEF \
  rawpath=rawdir$
#note that the shift is 4.7 pixels for object cal
#also, note that there appears to be an error in the WCS.

#------------make flat----------------
#first check flats
del phu.lis
gemextn n@flat.lis proc="expand" extname="SCI" > phu.lis
imstat @phu.lis >> GS-CAL20040615.log 
imstat @phu.lis
#first two are slightly off from the rest of the set
del flat.lis
gemlist S20040615S 186-194 > flat.lis
imdel rnS20040615S0186_flat,rnS20040615S0186_flat_bpm.pl
#not required to cut longslit data, but it avoids a large number of 
#"bad" pixels in the dq array
imdel rn@flat.lis
nsreduce n@flat.lis fl_nscut+ fl_nsappwav- fl_sky- fl_flat- 
#no darks are needed for such short exposure times (only add noise)
nsflat rn@flat.lis darks="" 
#flatfield is rnS20040615S0186_flat
#display rnS20040615S0186_flat[sci] 1

#-----------wavelength calibrate arc---------
imdel rn@arc.lis
nsreduce n@arc.lis fl_sky- fl_flat- 
imdel wrn@arc.lis
del GS-CAL20040615_database/idwrnS20040615S0195_SCI_1_
#For lower dispersion, this task often works well non-interactively, BUT 
#  ALWAYS CHECK RESULT-- you might need to do it interactively.
nswave rn@arc.lis 

#reduce tellurics
del phu.lis
gemextn n@telluric.lis proc="expand" index="0" > phu.lis
hselect @phu.lis "$I,UT" yes
#20 seconds will include neighbours for sky
imdel rn@telluric.lis
nsreduce n@telluric.lis skyrange=20 fl_flat+ flatim=rnS20040615S0186_flat \
  fl_vardq+

##################
#FYI:
#nssstack combines things at the same offset position (no shifting) 
#  and gives 1 output file per position
#nscomb (used below) shifts all data spatially to a common offset 
#  position and gives 1 combined output file
################## 
#nsstack telluric to provide input to nssdist
#need to use the 'n'prepared files (ie, before sky-subtraction) so 
#cut explicitly (or could call nsreduce with fl_nscut+ and a different
#output prefix)
imdel sn@telluric.lis
nscut n@telluric.lis
imdel snS20040615S0145_stack,snS20040615S0146_stack
nsstack sn@telluric.lis

#This creates a distortion correction in the spatial direction
#IMPORTANT: set firstcoord close to the position of the spectrum in the 
#first (left-most!) image
#(do this, for example, by displaying image...
#display snS20040615S0145_stack[sci,1] 1
#...and see x=329 at emission, more or less)
del GS-CAL20040615_database/idsnS20040615S0145_stack_sdist_SCI_1_
imdel snS20040615S0145_stack_sdist
nssdist snS20040615S0145_stack,snS20040615S0146_stack firstcoord=329

#call nstransform to straighten and dispersion correct, then nscomb 
#on reduced frames (skip nsstack)
imdel trn@telluric.lis
nstransform rn@telluric.lis lamp=wrnS20040615S0195 \
  sdist=snS20040615S0145_stack_sdist
#or, if we didn't want to correct for spatial distortion:
#nstransform rn@telluric.lis lamp=wrnS20040615S0195 

#combine 2D spectra 
imdel trnS20040615S0145_comb
nscomb trn@telluric.lis

#now extract to 1-D
imdel xtrnS20040615S0145_comb
nsextract trnS20040615S0145_comb

#final s/n ~50
#splot xtrnS20040615S0145_comb[sci,1]

#Cleanup: 
delete tmp*,*.lis
imdel nS20040615S????.fits
imdel snS20040615S????.fits
imdel rnS20040615S????.fits
imdel trnS20040615S????.fits

#Packing
#Run cutlog and create packlists (Gemini internal use)
#delete 20040615_GS-CAL20040615_obslog.txt
#gemcutlog ../../Obslogs/2004jun15_obslog.txt GS-CAL20040615 > 20040615_GS-CAL20040615_obslog.txt
#delete 20040615_GS-CAL20040615_pklist.txt
#gspklist 20040615_GS-CAL20040615_obslog.txt rawpath="/net/veronica/staging" xmlpath=""

